#include<assert.h>
#include<stdlib.h>
#include<stdio.h>
#include "Enumerate.h"
/*Enumerates all the permutations*/
void EnumAllPerms(int n, int *perm, int *next_avail, fperm_ptr UsePerm, void *x){
	int curr_ptr=n,i,find;

	/*This is the permutation we start-off with*/
	for(i=0;i<n;i++){
		perm[i] = i+1;
		next_avail[i] = 0;
	}

	while(curr_ptr >= 0){
		if(curr_ptr == n){
			UsePerm(perm,n,x);
			curr_ptr-=1;
		}
		/*Find the next smallest i > perm[curr_ptr]*/
		find=1;
		while((perm[curr_ptr]+find-1 < n) && 
			!(next_avail[perm[curr_ptr]+find-1])){
			find++;
		}
		if(!(perm[curr_ptr]+find-1 < n)){
			next_avail[perm[curr_ptr]-1] = 1;
			perm[curr_ptr]=0;
			curr_ptr--; continue;
		}
		next_avail[perm[curr_ptr]+find-1] = 0; 
		if(perm[curr_ptr] > 0){
			next_avail[perm[curr_ptr]-1] = 1;
		}
		perm[curr_ptr] = perm[curr_ptr]+find;
		curr_ptr++;
	}
}
/* Apply permutation, applies a permutation on a linear ordered 
 * objects, this is very general in the sense it just generates
 * a SWAP policy (i.e. just a sequence of SWAP instructions).
 * note the permutation is w.r.t to the order \{1,2,3,\ldots n\}
 *
 * TODO: Study Knuth's paper on permuting inplace.
 */
static int pick_nid[4096]={0};
void ApplyPermutation(int *perm,int n, void (*swap)(int,int,void *), void *x){ 
	int curr=0,nidx=0; 
	int cvalue=1; /*Left most unmatched permutation position*/ 
	int count=0; int i;
	static int token = 0; /*Avoid spending time in Initialization*/
	token++;
	assert(n < 4096);

	while(curr < n){
		pick_nid[curr] = token;
		if(cvalue != perm[curr]){
			swap(curr,perm[curr]-1,x);
			curr = perm[curr]-1;
		}else{
			/*Find the start of the next cycle*/
			while( nidx < n && 
				(pick_nid[nidx] == token || perm[nidx]==nidx+1)){
				nidx++;
			}
			/*Now make that curr*/
			curr = nidx; cvalue = nidx+1;
		}
	}
}
void SimpleSwap(int i, int j,void *x){
	int *a = (int *)x;
	int temp;

	temp = a[i]; a[i]=a[j]; a[j]=temp;
}


/*Simple Function to use the permutation in the array*/
static unsigned long long pcount=0; /*Make sure $log(n!)_2 < 64$*/
void* PrintPerm(int *perm, int dim, void *x){
	int i;
	pcount++;
	for(i=0; i< dim; i++){
		printf("%d ", perm[i]);
	}
	printf("\n");
}

#ifdef UNIT_TEST_ENUMERATE
int main(int argc,char **argv){
	int n,i;
	int *perm=NULL; 
	int *next_avail=NULL;
	int *a=NULL;

	a = (int *) malloc(sizeof(int)*9);
	perm = (int *) malloc(sizeof(int)*9);

	while(1){
		for(i=0;i<9;i++){
			a[i] = i+1;
		}
		printf("Please enter a permutation\n");
		for(i=0;i<9;i++){
			scanf("%d",&perm[i]);
			assert(perm[i] <= 9 && perm[i]>0);
		}
		ApplyPermutation(perm,9,SimpleSwap,a);

		printf("Result\n");
		for(i=0; i<9; i++){
			printf("%d ", a[i]);
		}

	}
	
#if 0
	fperm_ptr useperm = PrintPerm;
	if(argc < 2){
		fprintf(stderr,"USAGE: perm <integer n>");
		exit(1);
	}
	n = atoi(argv[1]);
	next_avail = (int *) malloc(sizeof(int )*n);
	perm = (int *) malloc(sizeof(int )*n);
	assert(next_avail && perm);
	EnumAllPerms(n,next_avail,perm,useperm);
	printf("!%d=%llu\n",n,pcount);
	free(next_avail); 
	free(perm);
#endif
}
#endif
