/*Hier-BLM: Solve the BLM problem hier
 * 
 * 09/12/2009 vamsik@engr.uconn.edu
 */
#include<stdlib.h>
#include<math.h>
#include<string.h>
#include<assert.h>
#include<stdio.h>
#include "hier_blm.h"
#include "Enumerate.h"
#include "CQueue.h"
unsigned long PDIM=3; /*Default dimension of the sub problem*/
unsigned long PDIM2=9; /*Default square*/
unsigned long **GDT=NULL;

unsigned long ipower(unsigned long level){
	static unsigned long prev_level = 0; 
	static unsigned long value=1;
	unsigned long pm=PDIM; unsigned long p3l=1;
	if(prev_level == level){
		return value;
	}
	prev_level = level;
	while(level){
		if(level & 1){
			p3l *= pm;
		}
		pm *=pm;
		level >>= 1;
	}
	value = p3l;
	return p3l;
}


#define SUBDIST(i,j,k) sp->SUB_DIST_TABLE[((i)*(PDIM*PDIM*2))+(j)*2+(k)]
/*Actual cost computation routine*/
unsigned long ComputeSubProblemCost(int *perm, SubProblem *sp){
	unsigned int i,j;
	unsigned long cost_x =0, cost_y=0;
	assert(PDIM > 1);
	/*Cost along the rows*/
	for(i=0; i<PDIM; i++){
		for(j=0; j<PDIM-1; j++){
			cost_x += SUBDIST(perm[j+i*PDIM]-1,perm[j+1+i*PDIM]-1,1);
		}
	}
	/*Cost along the cols*/
	for(i=0; i<PDIM-1; i++){
		for(j=0; j<PDIM; j++){
			cost_y += SUBDIST(perm[i*PDIM+j]-1,perm[i*PDIM+j+PDIM]-1,0);
		}
	}
	return cost_x+cost_y;
} 
/*Takes a sub-problem and computes the placement cost
 *of the squares based on the orientation -- i.e. L->R or T->B
 *
 *NOTE:This assumes the squares are laid-down in the row major
 *order.
 */
unsigned long SquareCost(SubProblem *sp, int i, int j, int o){
	unsigned long width = ipower(sp->level);
	unsigned long cost=0; unsigned int k;
	unsigned int *p = sp->probes;

	/*Simple Special case*/
	if(width == 1 || i==j){
		return GDT[p[i]-1][p[j]-1];
	}

	unsigned long width2 = width*width;
	unsigned int is = (unsigned int)i*width2;
	unsigned int js = (unsigned int)j*width2;
	if(o){ /*Left to right*/
		for(k=0; k<width; k++){
			cost += GDT[p[is+((k+1)*width)-1]-1][p[js+(k*width)]-1];
		}
	}else{ /*Top to bottom*/
	 	for(k=0; k<width; k++){
			cost += GDT[p[is+width2-width+k]-1][p[js+k]-1];
		}
	}
	return cost;
}
/*This swaps the sup-problem segments of the Squares
 *NOTE: please use this for merging sub-problems*/
void SwapSubProblemSegments(int i, int j, void *s){
	SubProblem *sp = (SubProblem *) s;
	unsigned long width = ipower(sp->level); /*This is the width used to swap*/
	unsigned int k;

	/*Swap two memory segments with-out extra buffers*/
	for(k=0; k<width; k++){
		sp->probes[i*width+k] = (sp->probes[i*width+k])^(sp->probes[j*width+k]);
		sp->probes[j*width+k] = (sp->probes[i*width+k])^(sp->probes[j*width+k]);
		sp->probes[i*width+k] = (sp->probes[i*width+k])^(sp->probes[j*width+k]);
	}
}
/* This swaps the Squares -- which are stored in the row major order,
 * within the sub-problem*/
void SwapSquares(int i, int j, void *s){
	SubProblem *sp = (SubProblem *) s;
	unsigned long width = ipower(sp->level);
	unsigned long width2 = width*width;
	unsigned int k;
	for(k=0; k<width2; k++){
		sp->probes[i*width2+k] = (sp->probes[i*width2+k])^(sp->probes[j*width2+k]);
		sp->probes[j*width2+k] = (sp->probes[i*width2+k])^(sp->probes[j*width2+k]);
		sp->probes[i*width2+k] = (sp->probes[i*width2+k])^(sp->probes[j*width2+k]);
	}
}

/*Call this after you apply the permutation.
 *
 * Note: Merging is nothing but permuting using 
 * a special permutation. Just create that special 
 * permutation and call the ApplyPermutation.
 *
 * Also its important to keep in mind that optimal 
 * permutation is applied then this is called.
 **/
void MergeSubProblems(SubProblem *sp){ 
	static int *s = NULL; unsigned int k,i,j,m;
	static unsigned int slevel = 0;
	unsigned long PDIM2 = PDIM*PDIM;
	unsigned long width = ipower(sp->level);

	if(sp->level != slevel){ /*This needs to be generated once per-level*/
		slevel = sp->level;
		if(s){
			free(s);
		}
		s = (int *) malloc(sizeof(int)*(PDIM2*width));
		assert(s);
		m=0;
		for(k=1; k<=PDIM; k++){
			for(i=1; i<=width; i++){
				for(j=1; j<=PDIM; j++){
					s[m++]  = i + (j-1)*width + (k-1)*PDIM*width;
				}
			}
		}
	}
	assert(sp->psize == (unsigned int) PDIM2*width*width);
	/*Apply permutation s to sp->probes*/
	ApplyPermutation(s,(int)(PDIM2*width),SwapSubProblemSegments,sp);
}

/*Create a Subproblem distance matrix*/
void CreateSubProblemDistanceMatrix(SubProblem *sp){
	/*This matrix is always of size $PDIM^2\times 2$*/
	unsigned long *sm = NULL; unsigned int i,j,k;
	unsigned long PDIM2 = PDIM*PDIM;

	/*can be accessed by (i*(2*PDIM)+j*PDIM+k) */
	sp->SUB_DIST_TABLE = (unsigned long *) malloc(sizeof(unsigned long)*
		(PDIM2*PDIM2*2));
	assert(sp->SUB_DIST_TABLE);
	sm = sp->SUB_DIST_TABLE;

	/*This matrix is un-symmetric*/
	for(i=0; i<PDIM2; i++){
		for(j=0; j<PDIM2; j++){
			for(k=0; k<2; k++){ /*0 is top-bottom, 1 is left-right*/
				/*Compute the value of SUB_DIST_TABLE[i][j][k]*/
				sm[i*(PDIM2*2)+j*2+k] = SquareCost(sp,i,j,k);
			}
		}
	}
}
/*Free a sub-problem*/
void FreeSubProblem(SubProblem *sp){
	/*Free probes*/
	free(sp->probes);
	free(sp->SUB_DIST_TABLE);
	if(sp->best_perm) free(sp->best_perm);
}

unsigned long HammingDistance(char *s1, char *s2);
unsigned long BuildGlobalDistanceTable(FILE *probe_file){
	unsigned long psize=0;
	long next_seek; unsigned int i,j;
	char s1[30], s2[30]; /*make this strictly > probe-size*/
	
	assert(fscanf(probe_file,"%lu",&psize) == 1);
	
	assert(psize > 1);

	/*Create the global distance table.*/
	GDT = (unsigned long **) malloc(sizeof(unsigned long *)*psize);
	for(i=0; i<psize; i++){
		GDT[i] = (unsigned long *) malloc(sizeof(unsigned long)*psize);
	}
	i=0;

	printf("Building distance table.....\n");
	for(i=0; i<psize; i++){
		/*read the string s[i]*/
		printf("processing | %u\r",i);
		printf("processing / %u\r",i);
		printf("processing - %u\r",i);
		printf("processing \\ %u\r",i); fflush(stdout);
		assert(fscanf(probe_file,"%25s",s1) >= 1);
		next_seek = ftell(probe_file); 
		for(j=i+1; j<psize; j++){
			assert(fscanf(probe_file,"%25s",s2)>=1);
			GDT[i][j] = HammingDistance(s1,s2);
			GDT[j][i] = GDT[i][j];
		}
		assert(!(fseek(probe_file,next_seek,SEEK_SET)));
	}
	fclose(probe_file);
	return psize;
}

unsigned long HammingDistance(char *s1, char *s2){
	unsigned long i=0;
	unsigned long hdistance=0;
	while(s1[i]!='\0' && s2[i]!='\0'){
		if(s1[i]!=s2[i]){
			hdistance++;
		}
		i++;
	}
	if(s1[i]=='\0'){
		while(s2[i++]!='\0'){
			hdistance++;
		}
	}else{
		while(s1[i++]!='\0'){
			hdistance++;
		}
	}
	return hdistance;
}
/*Finds the best permutation, NOTE: please set sp->best_cost = (unsigned long)-1;*/
/*also set sp->best_perm = NULL*/
void* FindBestPerm(int *perm, int n, void *s){
	SubProblem *sp = (SubProblem *)s;
	unsigned long cost = ComputeSubProblemCost(perm,sp);
	if(cost < sp->best_cost){
		sp->best_cost = cost;
		//printf("best cost = %lu \n",sp->best_cost);
		if(!sp->best_perm){
			sp->best_perm = (int *)malloc(sizeof(int)*n);
			assert(sp->best_perm);
		}
		assert(memcpy(sp->best_perm,perm,n*sizeof(int)) == sp->best_perm);
	}
}

/*Create Super Problem -- from PDIM^2 sub-problems*/
/*This is the general merge approach we can do better 
 *by using that fact that we always merge PDIM2 sub-problems
 *each of size 3^l
 */
SubProblem *CreateSuperProblem(SubProblem **sub_array, unsigned int n){
	SubProblem *super = (SubProblem *) malloc(sizeof(SubProblem));
	unsigned long super_psize=0; unsigned int i,psize=0;
	for(i=0; i<n; i++){
		super_psize += sub_array[i]->psize;
	}
	super->probes = (unsigned int *) malloc(super_psize*sizeof(unsigned int));
	super->level = sub_array[0]->level+1;
	super->psize = super_psize;
	assert(super->probes);
	for(i=0; i<n; i++){
		assert(memcpy(&(super->probes[psize]),sub_array[i]->probes,
			sizeof(unsigned int)*(sub_array[i]->psize)) == &(super->probes[psize]));
		psize += sub_array[i]->psize;
		FreeSubProblem(sub_array[i]);
	}
	assert(psize == super->psize);
	/*init it*/
	super->best_perm = NULL;
	super->best_cost = (unsigned int)-1;
	return super;
}

/*Computes the cost of the Initial permutation*/
unsigned long PlacementCost(int N, unsigned int *probes){
	unsigned int i,j;
	unsigned int dim = (unsigned int) sqrt(N);
	unsigned long cost_x=0,cost_y=0;

	printf("Computing the initial cost with dim=%u \n",dim);
	for(i=0; i<dim; i++){
		for(j=0; j<dim-1; j++){
			if(!probes){
				cost_x += GDT[j+i*dim][j+1+i*dim];
			}else{
				cost_x += GDT[probes[j+i*dim]-1][probes[j+1+i*dim]-1];
			}
		}
	}
	for(i=0; i<dim-1; i++){
		for(j=0; j<dim; j++){
			if(!probes){
				cost_y += GDT[i*dim+j][i*dim+j+dim];
			}else{
				cost_y += GDT[probes[i*dim+j]-1][probes[i*dim+j+dim]-1];
			}
		}
	}
	return cost_x + cost_y;
}

/*The main toplevel routine to solve the BLM problem*/
SubProblem *HierBLMSolve(int N){ /*Just the number of probes*/
	unsigned int i,j;
	SubProblem *scache = NULL; 
	SubProblem **mbuffer = NULL;
	int *permbuf=NULL,*nbuf=NULL;
	unsigned int midx=0;
	unsigned long total_cost=0;
	SubProblem *auxs=NULL;
	unsigned int *placement;
	/*circular queue to keep track of the sub-problems*/
	CQueue *cq = NULL;

	assert(GDT); /*Makesure that the global distance table is built*/
	assert(!(N%PDIM2)); /*Make sure the # of probes is a multiple of PDIM2*/


	mbuffer = (SubProblem **)malloc(sizeof(SubProblem *)*PDIM2);
	permbuf = (int *)malloc(sizeof(int)*PDIM2);
	nbuf = (int *)malloc(sizeof(int)*PDIM2);

	assert(mbuffer && permbuf && nbuf);
	
	/*Setup the circular queue*/
	cq = (CQueue *) malloc(sizeof(CQueue));
	InitializeCQ(cq,sizeof(SubProblem *));


	/*Now append all the sub-problems*/
	scache = (SubProblem *) malloc(sizeof(SubProblem)*(N/PDIM2));
	for(i=0; i<N/PDIM2; i++){
	 	(scache[i]).probes = (unsigned int *)malloc(sizeof(unsigned int)*PDIM2);
	 	for(j=0; j<PDIM2; j++){
			 (scache[i]).probes[j]=i*PDIM2+j+1;
	 	}
	 	(scache[i]).psize = PDIM2;
	 	(scache[i]).best_perm = NULL; 
		(scache[i]).level = 0;
	 	(scache[i]).best_cost = (unsigned int)-1;
		auxs = &(scache[i]);
	 	AppendCQ(cq,&auxs);
	}

	printf("INITIAL PLACEMENT COST: %lu \n",PlacementCost(N,NULL));
	total_cost = 0;
	while(DequeueCQ(cq,&auxs)){
		mbuffer[midx] = (SubProblem *)auxs;
		/*First Create the sub-problem distance matrix*/
		CreateSubProblemDistanceMatrix(auxs);
		/*Solve this sub-problem optimally*/
		auxs->best_perm = NULL; auxs->best_cost=(unsigned int)-1;
		EnumAllPerms(PDIM2,permbuf,nbuf,FindBestPerm,auxs);
		/*Apply the permutation*/
		ApplyPermutation(auxs->best_perm,PDIM2,SwapSquares,auxs);
		/*Reoder the sub-problem*/
		total_cost += auxs->best_cost;
		if(auxs->level){
			MergeSubProblems(auxs);
		}
		printf("best_cost = %lu, verified_cost = %lu \n",auxs->best_cost, PlacementCost(auxs->psize,auxs->probes));
		placement = auxs->probes;
		midx++;
		if(!(midx%PDIM2)){ /*Merge these subproblems*/
			auxs = CreateSuperProblem(mbuffer,PDIM2);
			printf("LEVEL=%lu , Cum-Cost=%lu\n",auxs->level,total_cost);
			AppendCQ(cq,&auxs);
			midx=0;
		}
	}
	printf("FINAL PLACEMENT COST: %lu\n",total_cost);
	/*Verify this*/
	printf("VERIFYING PLACEMENT COST: %lu\n",PlacementCost(N,placement));

	/*TODO: Freeup all the stuff*/
}

int main(int argc, char **argv){
	FILE *file1 = NULL; int *perm=NULL;
	int i; int *next_avail; unsigned long psize;
	if(argc != 3){
		fprintf(stderr,"USAGE: blm-solve <dim> <file-name>\n");
		exit(1);
	}
	/*Create a sub-problem*/
	file1 = fopen(argv[2],"r"); assert(file1);
	PDIM = atoi(argv[1]); 
	PDIM2 = PDIM*PDIM;
	printf("PDIM=%lu \n",PDIM);
	/*build GlobalDistanceTable */
	psize = BuildGlobalDistanceTable(file1);

	if(psize%PDIM2){
		fprintf(stderr,"ERROR: Input should be a multiple of %lu\n",PDIM2);
		exit(1);
	}

	HierBLMSolve(psize);
}

#ifdef UNIT_TEST_HIER_BLM_TEST_COST
/*UNIT-TEST1: Test the construction of SubProblem and 
 *how the SUB_DIST_TABLE is build
 */
int main(){
	/*Create a sub-problem with 9 probes*/
	sp->probes = (int *)malloc(sizeof(int)*81);
	perm = (int *)malloc(sizeof(int)*81);
	next_avail = (int *)malloc(sizeof(int)*81);
	sp->psize = 81;
	sp->sid=1;
	sp->level=1;
	sp->best_perm=NULL;


	/*Find the cost of simple permutation*/
	for(i=0;i<81;i++){
		sp->probes[i] = i;
	}

	perm[0]=9; perm[8]=1;
	for(i=1; i<8; i++){
		perm[i]=i+1;
	}



	/*Now lets create the sub-problem distance matrix*/
	CreateSubProblemDistanceMatrix(sp);
	printf("Running Merge Sub Problems Test.....\n");
	for(i=1; i<=81; i++){
		printf("%u ",sp->probes[i-1]);
		if(!(i%9)){
			printf("\n");
		}
	}
	/*Find the best permutation*/
	sp->best_perm = NULL; sp->best_cost = (unsigned long)-1;
	EnumAllPerms(9,perm,next_avail,FindBestPerm,sp);

	printf("The cost of this permutation is %lu \n",ComputeSubProblemCost(sp->best_perm,sp));

	/*Apply a permutation */
	printf("applying the optimal permutation\n");
	ApplyPermutation(sp->best_perm,9,SwapSquares,sp);
	for(i=1; i<=81; i++){
		printf("%u ",sp->probes[i-1]);
		if(!(i%9)){
			printf("\n");
		}
	}

	printf("Placement after merging the sub problems\n");
	MergeSubProblems(sp);
	unsigned int l=0,k; 
	unsigned int lwidth=0;
	for(i=1; i<=81; i++){
		lwidth+=printf("%u ",sp->probes[i-1]);
		if(!(i%(ipower(sp->level)))){
			lwidth+=printf("| ");
		}
		if(!(i%9)){
			l++;
			if(!(l%3)){
				printf("\n");
				for(k=0;k<lwidth;k++){
					printf("=");
				}
				l=0;
			}
			printf("\n"); lwidth=0;
		}
	}
	FreeSubProblem(sp);
}
#endif
